#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) Copyright by Pierre-Henri Wuillemin, UPMC,
# 2017  (pierre-henri.wuillemin@lip6.fr)

# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and
# without fee or royalty is hereby granted, provided
# that the above copyright notice appear in all copies
# and that both that copyright notice and this permission
# notice appear in supporting documentation or portions
# thereof, including modifications, that you make.

# THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
# OR PERFORMANCE OF THIS SOFTWARE!

"""
Samples generation w.r.t to a probability distribution
represented by a Bayesian network.
"""

from __future__ import print_function

import csv
import math
import os
import random
import sys

import pyAgrum as gum

from ._utils.progress_bar import ProgressBar
from ._utils.pyAgrum_header import pyAgrum_header


class CSVGenerator:
  """
  Class for samples generation w.r.t to a probability distribution
  represented by a Bayesian network.
  """

  def __init__(self):
    print("** Class 'CSVGenerator' is is deprecated in pyAgrum>0.12.0. Please use gum.generateCSV instead.")
    self._parents = {}
    self._probas = {}

  @staticmethod
  def draw(tab):
    """
    draw a value using tab as probability table.

    Parameters
    ----------
    tab : list
      a probability table

    Returns
    -------
    tuple
      the couple (i,proba)
    """
    res = len(tab) - 1

    t = random.random()
    for i in range(len(tab)):
      if t < tab[i]:
        res = i
        break
      t -= tab[i]

    return (res, tab[res])

  @staticmethod
  def nameAndParents(bn, n):
    """
    Compute a list of parents for node n in BN bn.

    Parameters
    ----------
    bn : pyAgrum.BayesNet
      a Bayesian network
    n :
      (int) a node id
    n :
      (str) a node name

    Returns
    -------
    tuple
      a couple of name of n and list of parents names

    Raises
    ------
    gum.IndexError
      If the node is not in the Bayesian network
    """
    l = bn.cpt(n).var_names
    node = l.pop()
    # l is the list of var names
    return node, l  # map(bn.idFromName,l)

  def cachingnameAndParents(self, bn, n):
    """
    Compute a list of parents for node n in BN bn.

    Parameters
    ----------
    bn : pyAgrum.BayesNet
      a Bayesian network
    n :
      (int) a node id
    n :
      (str) a node name

    Returns
    -------
    tuple
      a couple of name of n and list of parents names

    """
    if not n in self._parents:
      self._parents[n] = CSVGenerator.nameAndParents(bn, n)
    return self._parents[n]

  def caching_probas(self, bn, node_id, n, par):
    """
    Parameters
    ----------
    bn : pyAgrum.BayesNet
      a Bayesian network
    node_id : int
      a node id
    n : int
      a node id
    par : list
      the node's parents

    Returns
    -------
    list
      the node's probabilities

    """
    if not n in self._probas:
      self._probas[n] = {}
    current = self._probas[n]

    for k in par:
      val_k = self._sample[k]
      if not val_k in current:
        current[val_k] = {}
      current = current[val_k]

    if not 'p' in current:
      current['p'] = list(bn.cpt(node_id)[dict([(p, self._sample[p]) for p in par])])

    return current['p']

  def newSample(self, bn, seq):
    """
    Generate a sample w.r.t to the bn using the variable sequence seq (topological order)

    Parameters
    ----------
    bn : pyAgrum.BayesNet
      a Bayesian network
    seq : list
      a variable sequence

    Returns
    -------
    tuple
      the coule (sample,log2-likelihood)
    """
    self._sample = {}
    LL = 0
    for node_id in seq:
      (nod, par) = self.cachingnameAndParents(bn, node_id)
      (self._sample[nod], p) = CSVGenerator.draw(self.caching_probas(bn, node_id, nod, par))
      LL += math.log(p, 2)

    return (self._sample, LL)

  def proceed(self, name_in, name_out, n, visible, with_labels):
    """
    From the file name_in (BN file), generate n samples and save them in name_out

    Parameters
    ----------
    name_in : str
      a file name
    name_out : str
      the output file
    n : int
      the number of samples
    visible : bool
      indicate if a progress bar should be displayed
    with_labels : bool
      indicate if values should be labelled or not

    Returns
    -------
    double
      the log2-likelihood of the n samples database
    """

    if isinstance(name_in, str):
      bn = gum.loadBN(name_in)
    else:
      bn = name_in

    seq = bn.topologicalOrder()
    writer = csv.writer(open(name_out, 'w',newline=''))

    if visible:
      sys.stdout.flush()

    titles = [bn.variable(item).name() for item in seq]
    writer.writerow(titles)

    if visible:
      prog = ProgressBar(name_out + ' : ', 0, n, 77, mode='static', char='#')
      prog.display()

    LL = 0
    for i in range(n):
      (cas, ll) = self.newSample(bn, seq)
      LL += ll

      if with_labels:
        row = [bn.variable(item).label(cas[bn.variable(item).name()]) for item in seq]
      else:
        row = [cas[bn.variable(item).name()] for item in seq]

      writer.writerow(row)

      if visible:
        prog.increment_amount()
        prog.display()

    if visible:
      print()
      print("Log2-Likelihood : {0}".format(LL))
      print()

    return LL

def generateCSV(name_in, name_out, n, visible=False, with_labels=True):
  """
  From the file name_in (BN file), generate n samples and save them in name_out

  Parameters
  ----------
  name_in : str
    a file name
  name_out : str
    the output file
  n : int
    the number of samples
  visible : bool
    indicate if a progress bar should be displayed
  with_labels : bool
    indicate if values should be labelled or not

  Returns
  -------
  double
    the log2-likelihood of the n samples database

  """
  print("** The function 'gum.lib.bn2csv.generateCSV' is deprecated in pyAgrum>0.12.0. Please use gum.generateCSV instead.")
  g = CSVGenerator()
  return g.proceed(name_in, name_out, n, visible, with_labels)

def module_help(exit_value=1):
  """
  defines help viewed if args are not OK on command line, and exit with exit_value
  """
  print(os.path.basename(sys.argv[0]), "src.{" + gum.availableBNExts() + "} nbr [dst.csv]")
  sys.exit(exit_value)
print("** The library 'gum.lib.bn2csv' is deprecated in pyAgrum>0.12.0. Please use gum.generateCSV instead.")

if __name__ == "__main__":
  pyAgrum_header("2011-2017")

  if len(sys.argv) < 3:
    module_help()

  bn_name = sys.argv[1]
  try:
    n = int(sys.argv[2])
  except ValueError:
    module_help()

  if len(sys.argv) < 4:
    base, ext = os.path.splitext(bn_name)
    csv_name = base + '.csv'
  else:
    csv_name = sys.argv[3]
    base, ext = os.path.splitext(csv_name)
    if ext != '.csv':
      csv_name += '.csv'

  generateCSV(bn_name, csv_name, n, True)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) Copyright by Pierre-Henri Wuillemin, UPMC, 2017
# (pierre-henri.wuillemin@lip6.fr)

# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and
# without fee or royalty is hereby granted, provided
# that the above copyright notice appear in all copies
# and that both that copyright notice and this permission
# notice appear in supporting documentation or portions
# thereof, including modifications, that you make.

# THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
# OR PERFORMANCE OF THIS SOFTWARE!

from __future__ import print_function

import csv
import math
import os
import sys

import pyAgrum as gum

from ._utils.progress_bar import ProgressBar
from ._utils.pyAgrum_header import pyAgrum_header


def lines_count(filename):
  """ count lines in a file
  """
  numlines = 0
  for line in open(filename): numlines += 1

  return numlines


def checkCompatibility(bn, fields, csv_name):
  """check if variables of the bn are in the fields

  if not : return None
  if compatibilty : return a list of position for variables in fields
  """
  res = {}
  isOK = True
  for field in bn.names():
    if not field in fields:
      print("** field '{0}' is missing in {1}".format(field, csv_name))
      isOK = False
    else:
      res[bn.idFromName(field)] = fields[field]

  if not isOK:
    res = None

  return res


def computeScores(bn_name, csv_name, visible=False, transforme_label=None):
  if isinstance(bn_name, str):
    bn = gum.loadBN(bn_name)
  else:
    bn = bn_name

  nbr_lines = lines_count(csv_name) - 1

  csvfile = open(csv_name, "r")
  dialect = csv.Sniffer().sniff(csvfile.read(1024))
  csvfile.seek(0)

  batchReader = csv.reader(open(csv_name, 'r'), dialect)

  titre = next(batchReader)
  fields = {}
  for i, nom in enumerate(titre):
    fields[nom] = i

  positions = checkCompatibility(bn, fields, csv_name)
  if positions is None:
    sys.exit(1)

  inst = bn.completeInstantiation()

  if visible:
    prog = ProgressBar(csv_name + ' : ', 0, nbr_lines, 77, mode='static', char='#')
    prog.display()

  nbr_insignificant = 0
  num_ligne = 0
  likelihood = 0.0
  for data in batchReader:
    num_ligne += 1

    for i in range(inst.nbrDim()):
      try:
        inst.chgVal(i, getNumLabel(inst, i, data[positions[i]], transforme_label))
      except gum.OutOfBounds:
        print("out of bounds", i, positions[i], data[positions[i]], inst.variable(i))

    p = bn.jointProbability(inst)
    if p == 0.0:
      print(str(num_ligne) + ":" + str(inst))
      nbr_insignificant += 1
    else:
      likelihood += math.log(p, 2)
    if visible:
      prog.increment_amount()
      prog.display()

  if visible:
    print

  nbr_arcs = 1.0 * bn.sizeArcs()
  dim = 1.0 * bn.dim()

  aic = likelihood - dim
  aicc = 2 * aic - 2 * dim * (dim + 1) / (nbr_lines - dim + 1) if (nbr_lines - dim + 1 > 0) else "undefined"
  bic = likelihood - dim * math.log(nbr_lines, 2)
  mdl = likelihood - nbr_arcs * math.log(nbr_lines, 2) - 32 * dim  # 32=nbr bits for a params

  return ((nbr_lines - nbr_insignificant) * 100.0 / nbr_lines,
          {'likelihood': likelihood, 'aic': aic, 'aicc': aicc, 'bic': bic, 'mdl': mdl})


def module_help(exit_value=1):
  """
  defines help viewed if args are not OK on command line, and exit with exit_value
  """
  print(os.path.basename(sys.argv[0]), "src.{" + gum.availableBNExts() + "} [data[.csv]]")
  sys.exit(exit_value)


def getNumLabel(inst, i, label, transforme_label):
  if transforme_label is not None:
    label = transforme_label(label)

  if label.isdigit():  # an indice
    return int(label)
  else:
    return inst.variable(i)[label]


def stringify(s):
  return '"' + s + '"'


if __name__ == "__main__":
  pyAgrum_header("2011-17")

  if len(sys.argv) < 2:
    # module_help()
    bn_name = "../resources/alarm.dsl"
    csv_name = "alarm.csv"
  else:
    bn_name = sys.argv[1]

    if len(sys.argv) < 3:
      base, ext = os.path.splitext(bn_name)
      csv_name = base + '.csv'
    else:
      csv_name = sys.argv[2]
      base, ext = os.path.splitext(csv_name)

  print('"{0}" vs "{1}"'.format(bn_name, csv_name))
  print()
  (nbr, LL) = computeScores(bn_name, csv_name, visible=True)
  print()
  print('{0}% of base is significant.\nscores : {1}'.format(nbr, LL))



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) Copyright by Pierre-Henri Wuillemin, UPMC, 2017
# (pierre-henri.wuillemin@lip6.fr)

# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and
# without fee or royalty is hereby granted, provided
# that the above copyright notice appear in all copies
# and that both that copyright notice and this permission
# notice appear in supporting documentation or portions
# thereof, including modifications, that you make.

# THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
# OR PERFORMANCE OF THIS SOFTWARE!

from __future__ import print_function

import sys

import os
import pyAgrum as gum
import pydotplus as dot

from ._utils.pyAgrum_header import pyAgrum_header


def parents_name(bn, n):
  """
  Parameters
  ----------
  bn : pyAgrum.BayesNet
  	a Bayesian network
  n :
  	(str) the name of the node
  n :
	(int) the id of the node

  Returns
  -------
  map
  	a list of name of parents of node n

  Raises
  ------
  gum.IndexError
    If the node is not in the Bayesian network	
  """
  return map(lambda n: bn.variable(n).name(), bn.parents(nodeId(bn, n)))
  return l


def nodeId(bn, n):
  """
  Parameters
  ----------
  bn : pyAgrum.BayesNet
  	a Bayesian network
  n : str
  	the name of the node

  Returns
  -------
  int
  	the id of the node

  Raises
  ------
  gum.IndexError
    If the node is not in the Bayesian network	
  """
  if type(n) == str:
    return bn.idFromName(n)
  else:
    return n


DELTA_ERROR = 1e-6


def compareBNVariables(b1, b2):
  """
  Parameters
  ----------
  bn1 : pyAgrum.BayesNet
  	a Bayesian network
  bn2 : pyAgrum.BayesNet
  	another Bayesian network

  Returns
  -------
  str
  	'OK' if BN are composed of the same variables, indicates a non-existing variable otherwise

  """
  for i in range(b1.size()):
    try:
      v1 = b1.variable(i)
      v2 = b2.variable(b2.idFromName(v1.name()))
      if (v2.domainSize() != v1.domainSize()):
        return v1.name() + " has not the same domain size in the two bns"
    except IndexError:
      return b1.variable(i).name() + " does not exist in " + b2.property('name')

    for i in range(b2.size()):
      try:
        v2 = b2.variable(i)
        v1 = b1.variable(b1.idFromName(v2.name()))
        if (v2.domainSize() != v1.domainSize()):
          return v2.name() + " has not the same domain size in the two bns"
      except IndexError:
        return b2.variable(i).name() + " does not exist in " + b1.property('name')

  return "OK"


def compareBNParents(b1, b2):
  """
  Parameters
  ----------
  b1 : pyAgrum.BayesNet
  	a Bayesian network
  b2 : pyAgrum.BayesNet
  	another Bayesian network

  Returns
  -------
  str
  	'OK' if b2 have (at least) the same variable as b1 and their parents are the same.

  """
  for i in range(b1.size()):
    id1 = i
    id2 = b2.idFromName(b1.variable(id1).name())

    if not set(parents_name(b1, id1)) == set(parents_name(b2, id2)):
      return b1.variable(id1).name() + " has not the same parents in the two bns"

  return "OK"


def compareCPT(b1, cpt1, b2, cpt2):
  """
  Parameters
  ----------
  b1 : pyAgrum.BayesNet
  	a Bayesian network
  cpt1 : pyAgrum.Potential
  	one of b1's cpts
  b2 : pyAgrum.BayesNet
  	another Bayesian network
  cpt2 : pyAgrum.Potential
  	one of b2's cpts

  Returns
  -------
  str
  	'OK' if CPTs are the same

  Raises
  ------
  gum.KeyError
    If cpts are not from the same variable
  """

  dico2 = {}
  for i in range(cpt2.nbrDim()):
    dico2[cpt2.variable(i).name()] = i

  I1 = gum.Instantiation(cpt1)
  I2 = gum.Instantiation(cpt2)
  I1.setFirst()
  while not I1.end():
    for i in range(I1.nbrDim()):
      I2.chgVal(dico2[I1.variable(i).name()], I1.val(i))

    if abs(cpt1.get(I1) - cpt2.get(I2)) > DELTA_ERROR:
      return "For " + cpt1.variable(0).name() + ", cpt[" + str(I1) + "] (=" + str(cpt1.get(I1)) + ", in " + b1.property(
          'name') + ") !=cpt2[" + str(I2) + "] (=" + str(cpt2.get(I2)) + ", in " + b2.property('name') + ")"
    I1 += 1
  return "OK"


def compareBNCPT(b1, b2):
  """
  Parameters
  ----------
  b1 : pyAgrum.BayesNet
  	a Bayesian network
  b2 : pyAgrum.BayesNet
  	another Bayesian network

  Returns
  -------
  str
  	'OK' if b2 have (at least) the same variable as b1 and their cpts are the same
  """
  for i in range(b1.size()):
    res = compareCPT(b1, b1.cpt(i), b2, b2.cpt(b2.idFromName(b1.variable(i).name())))
    if res != "OK":
      return res
  return "OK"


def compareBN(name1, name2):
  """
  Parameters
  ----------
  name1 : str
  	a BN filename
  name2 : str
  	another BN filename

  Returns
  -------
  str
  	"OK" if bn are the same, a description of the error otherwise

  """
  if isinstance(name1, str):
    b1 = gum.loadBN(name1)
    b1.setProperty('name', '"' + os.path.basename(b1.property('name') + '"'))
  else:
    b1 = name1

  if isinstance(name2, str):
    b2 = gum.loadBN(name2)
    b2.setProperty('name', '"' + os.path.basename(b2.property('name') + '"'))
  else:
    b2 = name2

  ret = compareBNVariables(b1, b2)
  if ret != "OK":
    return ret

  ret = compareBNParents(b1, b2)
  if ret != "OK":
    return ret

  ret = compareBNCPT(b1, b2)
  return ret


def graphicalBNDiff(bn1, bn2):
  """ Return a pydotplus graph that compares the arcs of bn1 (reference) with those of bn2.
  full black line: the arc is common for both
  full red line: the arc is common but inverted in bn2
  dotted black line: the arc is added in bn2
  dotted red line: the arc is removed in bn2

  :param BayesNet bn1: referent model for the comparison
  :param BayesNet bn2: bn compared to the referent model
  :return: the result dot graph
  """

  if set([i for i in bn1.names()]) == set([i for i in bn2.names()]):

    graph = dot.Dot(graph_type='digraph', bgcolor="transparent")

    # Nodes
    for n in bn1.names():
      node = dot.Node('"' + n + '"', style="filled",
                      bgcol="#444444",
                      fgcol="#FFFFFF",
                      tooltip='"({0}) {1}{2}"'.format(bn1.idFromName(n), n, ""))
      graph.add_node(node)

    # Arcs
    for n1 in bn2.names():
      for n2 in bn2.names():
        id1 = bn2.idFromName(n1)
        id2 = bn2.idFromName(n2)
        ID1 = bn1.idFromName(n1)
        ID2 = bn1.idFromName(n2)

        if bn1.dag().existsArc(ID1, ID2) and bn2.dag().existsArc(id1, id2):
          # If present in both
          edge = dot.Edge('"' + n1 + '"', '"' + n2 + '"')
          graph.add_edge(edge)
        elif bn1.dag().existsArc(ID1, ID2) and bn2.dag().existsArc(id2, id1):
          # If inversed in bn2
          edge = dot.Edge('"' + n2 + '"', '"' + n1 + '"')
          edge.set_color("red")
          graph.add_edge(edge)
        elif bn1.dag().existsArc(ID1, ID2):
          # If removed from bn2
          edge = dot.Edge('"' + n1 + '"', '"' + n2 + '"', style='dashed')
          edge.set_color("red")
          graph.add_edge(edge)
        elif not bn1.dag().existsArc(ID1, ID2) and not bn1.dag().existsArc(ID2, ID1) and bn2.dag().existsArc(id1,
                                                                                                             id2):
          # If added to bn2
          edge = dot.Edge('"' + n1 + '"', '"' + n2 + '"', style='dashed')
          graph.add_edge(edge)

  else:
    raise ValueError('Not the same variables in the BNs')

  return (graph)


def module_help(exit_value=1):
  """
  defines help viewed if args are not OK on command line, and exit with exit_value
  """
  print(os.path.basename(sys.argv[0]), "bn1.{" + gum.availableBNExts() + "} bn2.{" + gum.availableBNExts() + "}")
  sys.exit(exit_value)


if __name__ == "__main__":
  pyAgrum_header("2011-17")

  if len(sys.argv) != 3:
    module_help()

  print(compareBN(sys.argv[1], sys.argv[2]))



# - * - coding : utf - 8 - * -
"""
Deprecated for older pyAgrum
"""
from .pyAgrum import GibbsSampling, Potential
from .pyAgrum import Arc, Edge, DiGraph, UndiGraph, MixedGraph, DAG, CliqueGraph
from .pyAgrum import BayesNet, EssentialGraph, MarkovBlanket


def GibbsInference(bn):
  """
  Deprecated class. Use pyAgrum.GibbsSampling instead.
  """
  print("** pyAgrum.GibbsInference is deprecated in pyAgrum>0.12.0.")
  print("** A pyAgrum.GibbsSampling has been created.")
  return GibbsSampling(bn)


def _addDeprecatedMethods():
  def deprecated_populate(pot, v):
    """
    Deprecated method in pyAgrum>0.12.0. See fillWith instead.
    """
    print("** pyAgrum.Potential.populate() is deprecated in pyAgrum>0.12.0.")
    print("** pyAgrum.Potential.fillWith() will be called instead.")
    return pot.fillWith(v)
  Potential.populate = deprecated_populate

  def deprecated_fill(pot, v):
    """
    Deprecated method in pyAgrum>0.12.0. See fillWith instead.
    """
    print("** pyAgrum.Potential.fill() is deprecated in pyAgrum>0.12.0.")
    print("** pyAgrum.Potential.fillWith() will be called instead.")
    return pot.fillWith(v)
  Potential.fill = deprecated_fill

  def deprecated_ids(gr):
    """
    Deprecated method in pyAgrum>0.12.0. See nodes instead.
    """
    print("** pyAgrum.*.ids() is deprecated in pyAgrum>0.12.0.")
    print("** pyAgrum.*.nodes() will be called instead.")
    return gr.nodes()
  DiGraph.ids=deprecated_ids # and subclasses
  UndiGraph.ids=deprecated_ids # and subclasses
  EssentialGraph.ids=deprecated_ids
  BayesNet.ids=deprecated_ids

#from .pyAgrum import BayesNet, EssentialGraph, MarkovBlanket



_addDeprecatedMethods()



# -*- coding: utf-8 -*-
"""
`pyAgrum <http://agrum.org>`_ is a Python wrapper for the C++ `aGrUM <https://agrum.lip6.fr>`_ library.
It provides a high-level interface to the C++ part of aGrUM allowing to create, manage and perform efficient computations with Bayesian Networks.

The module is generated using the `SWIG <https://www.swig.org>`_ interface generator. Custom-written code was added to make the interface more friendly.

pyAgrum includes :
  * API documentation <http://www-desir.lip6.fr/~phw/aGrUM/docs/last/pyAgrum/>,
  * examples as `notebooks <http://www-desir.lip6.fr/~phw/aGrUM/docs/last/notebooks/01-tutorial.ipynb.html>`_,
  * a `website <http://agrum.org>`_.
"""

# aGrum Licence (GPL)
# -------------------
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU General Public License as published by  *
#*   the Free Software Foundation; either version 2 of the License, or     *
#*   (at your option) any later version.                                   *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU General Public License for more details.                          *
#*                                                                         *
#*   You should have received a copy of the GNU General Public License     *
#*   along with this program; if not, write to the                         *
#*   Free Software Foundation, Inc.,                                       *
#*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

__version__ = '0.13.6'
__license__ = __doc__
__project_url__ = 'http://agrum.org'

import numpy as np
from .functions import *
from .deprecated import *

# selection of imports extracted from dir(.pyAgrum)
from .pyAgrum import statsObj
from .pyAgrum import Arc, Edge, DiGraph, UndiGraph, MixedGraph, DAG, CliqueGraph
from .pyAgrum import BayesNet, EssentialGraph, MarkovBlanket
from .pyAgrum import DiscretizedVariable, LabelizedVariable, RangeVariable, DiscreteVariable
from .pyAgrum import Potential, Instantiation, Potential
from .pyAgrum import BruteForceKL, GibbsKL, StructuralComparator
from .pyAgrum import LazyPropagation, ShaferShenoyInference, VariableElimination
from .pyAgrum import LoopyBeliefPropagation, GibbsSampling, MonteCarloSampling, ImportanceSampling, WeightedSampling
from .pyAgrum import LoopyImportanceSampling,LoopyGibbsSampling,LoopyWeightedSampling,LoopyMonteCarloSampling
from .pyAgrum import PythonApproximationListener, PythonBNListener, PythonLoadListener, PythonDatabaseGeneratorListener
from .pyAgrum import BNGenerator, IDGenerator, JTGenerator
from .pyAgrum import BNLearner
from .pyAgrum import InfluenceDiagram, InfluenceDiagramInference
from .pyAgrum import CredalNet, CNMonteCarloSampling, CNLoopyPropagation
from .pyAgrum import PRMexplorer

from .pyAgrum import ApproximationScheme

from .pyAgrum import initRandom, randomProba, randomDistribution, randomGeneratorSeed, getRandomGenerator, randomValue
from .pyAgrum import isOMP, setNumberOfThreads, getThreadNumber, getNumberOfRunningThreads, getNumberOfLogicalProcessors, getMaxNumberOfThreads, getDynamicThreadsNumber, setDynamicThreadsNumber

from .pyAgrum import getNestedParallelism, setNestedParallelism

from .pyAgrum import DefaultInLabel, DuplicateElement, DuplicateLabel, EmptyBSTree, EmptySet, GumException, FatalError, FormatNotFound, GraphError, IOError, IdError, InvalidArc, InvalidArgument, InvalidArgumentsNumber, InvalidDirectedCycle, InvalidEdge, InvalidNode, MissingValueInDatabase, MissingVariableInDatabase, NoChild, NoNeighbour, NoParent, NotFound, NullElement, OperationNotAllowed, OutOfBounds, OutOfLowerBound, OutOfUpperBound, ReferenceError, SizeError, SyntaxError, UndefinedElement, UndefinedIteratorKey, UndefinedIteratorValue, UnknownLabelInDatabase

# selection of imports extracted from dir(pyAgrum)
__all__=[
  '__version__','__license__','__project_url__',        # here
  'about','availableBNExts','loadBN','saveBN','loadID','getPosterior', # in fuctions.py
  'statsObj',

  'Arc','Edge','DiGraph','UndiGraph','MixedGraph','DAG','CliqueGraph',
  'BayesNet','EssentialGraph'
  'DiscretizedVariable','LabelizedVariable','RangeVariable','DiscreteVariable',
  'Potential','Instantiation','Potential',
  'BruteForceKL','GibbsKL','StructuralComparator',
  'LoopyBeliefPropagation','GibbsSampling','MonteCarloSampling', 'ImportanceSampling', 'WeightedSampling',
  'LoopyImportanceSampling','LoopyGibbsSampling','LoopyWeightedSampling','LoopyMonteCarloSampling'
  'LazyPropagation','ShaferShenoyInference','VariableElimination',
  'PythonApproximationListener','PythonBNListener','PythonLoadListener','PythonDatabaseGeneratorListener',
  'BNGenerator','IDGenerator','JTGenerator',
  'BNLearner',
  'BNDatabaseGenerator',
  'InfluenceDiagram','InfluenceDiagramInference',
  'CredalNet','CNMonteCarloSampling','CNLoopyPropagation',
  'PRMexplorer',

  'ApproximationScheme',

  'initRandom','randomProba','randomDistribution',
  'isOMP','setNumberOfThreads','getNumberOfLogicalProcessors','getMaxNumberOfThreads',

  'initRandom','randomProba','randomDistribution','randomGeneratorSeed','getRandomGenerator','randomValue',
  'isOMP','setNumberOfThreads','getThreadNumber','getNumberOfRunningThreads','getNumberOfLogicalProcessors','getMaxNumberOfThreads','getDynamicThreadsNumber','setDynamicThreadsNumber',

  'getNestedParallelism', 'setNestedParallelism',

  'DefaultInLabel','DuplicateElement','DuplicateLabel','EmptyBSTree','EmptySet','GumException','FatalError','FormatNotFound','GraphError','IOError','IdError','InvalidArc','InvalidArgument','InvalidArgumentsNumber','InvalidDirectedCycle','InvalidEdge','InvalidNode','MissingValueInDatabase','MissingVariableInDatabase', 'NoChild','NoNeighbour','NoParent','NotFound','NullElement','OperationNotAllowed','OutOfBounds','OutOfLowerBound','OutOfUpperBound','ReferenceError','SizeError','SyntaxError','UndefinedElement','UndefinedIteratorKey','UndefinedIteratorValue','UnknownLabelInDatabase'
  ]

# seed is chosen randomly :)
initRandom(0)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) Copyright by Pierre-Henri Wuillemin, UPMC, 2017
# (pierre-henri.wuillemin@lip6.fr)

# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and
# without fee or royalty is hereby granted, provided
# that the above copyright notice appear in all copies
# and that both that copyright notice and this permission
# notice appear in supporting documentation or portions
# thereof, including modifications, that you make.

# THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
# OR PERFORMANCE OF THIS SOFTWARE!

from __future__ import print_function

import csv
import os
import sys

import pyAgrum as gum

from ._utils.progress_bar import ProgressBar
from ._utils.pyAgrum_header import pyAgrum_header


def _lines_count(filename):
  """
  Parameters
  ----------
  filename : str
    a filename

  Returns
  -------
  int
    the number of lines in the file
  """
  numlines = 0
  for line in open(filename): numlines += 1

  return numlines


def _checkCompatibility(bn, fields, csv_name):
  """
  check if variables of the bn are in the fields

  Parameters
  ----------
  bn : pyAgrum.BayesNet
    a bayesian network
  fields : list
    a list of fields
  csv_name : str
    the name of the csv file

  Returns
  -------
  list
    a list of position for variables in fiels, None otherwise.
  """
  res = {}
  isOK = True
  for field in bn.names():
    if not field in fields:
      print("** field '{0}' is missing in {1}".format(field, csv_name))
      isOK = False
    else:
      res[bn.idFromName(field)] = fields[field]

  if not isOK:
    res = None

  return res


def _computeAUC(points):
  """
  Given a set of points drawing a ROC curve, compute the AUC value

  Parameters
  ----------
  points : set
    a set of points drawing a ROC curve

  Returns
  -------
  double
    the AUC value

  """
  # computes the integral from 0 to 1
  somme = 0
  for i in range(1, len(points)):
    somme += (points[i - 1][0] - points[i][0]) * points[i][1]

  return -somme


def __computeROCpoints(bn, csv_name, target, label, visible=False,with_labels=True):
  """
  Compute the ROC curve points.

  Parameters
  ----------
  bn : pyAgrum.BayesNet
    a bayesian network
  csv_name : str
    a csv filename
  target : str
    the target
  label : str
    the target's label
  visible : bool
    indicates if the resulting curve must be printed

  Returns
  -------
  tuple
    (res, totalP, totalN, idLabel)

  """
  idTarget = bn.idFromName(target)
  label=str(label)

  if not with_labels:
      idLabel = -1
      for i in range(bn.variable(idTarget).domainSize()):
        if bn.variable(idTarget).label(i) == label:
          idLabel = i
          break;
      assert (idLabel >= 0)

  engine = gum.LazyPropagation(bn)

  nbr_lines = _lines_count(csv_name) - 1

  if (sys.version_info >= (3, 0)):  # python 3
    csvfile = open(csv_name, "r")
    dialect = csv.Sniffer().sniff(csvfile.readline())
    csvfile.seek(0)
    batchReader = csv.reader(open(csv_name, 'r'), dialect)
    titre = batchReader.__next__()
  else:  # python2
    csvfile = open(csv_name, "rb")
    dialect = csv.Sniffer().sniff(csvfile.readline())
    csvfile.seek(0)
    batchReader = csv.reader(open(csv_name, 'rb'), dialect)
    titre = batchReader.next()
  fields = {}
  for i, nom in enumerate(titre):
    fields[nom] = i

  positions = _checkCompatibility(bn, fields, csv_name)

  if positions is None:
    sys.exit(1)

  if visible:
    prog = ProgressBar(csv_name + ' : ', 0, nbr_lines, 77, mode='static', char='#')
    prog.display()

  totalP = 0
  totalN = 0
  res = []
  for data in batchReader:
    if with_labels:
      if str(data[positions[idTarget]])==label:
        totalP += 1
      else:
        totalN += 1
    else:
      idTargetLine = int(data[positions[idTarget]])
      if idTargetLine == idLabel:
        totalP += 1
      else:
        totalN += 1

    engine.eraseAllEvidence()
    e = {}
    for var in bn.names():
      if not var.__eq__(target):
        if with_labels:
            e[var] = str(data[positions[bn.idFromName(var)]])
        else:
            e[var] = bn.variableFromName(var).label(int(data[positions[bn.idFromName(var)]]))

    try:
      engine.setEvidence(e)
      engine.makeInference()
      px = engine.posterior(idTarget)[{target: label}]
      if with_labels:
        res.append((px, str(data[positions[idTarget]])))
      else:
        res.append((px, int(data[positions[idTarget]])))
    except gum.OutOfBounds as err:
      print(err)
      print("erreur : " + str(e))

    if visible:
      prog.increment_amount()
      prog.display()

  if visible:
    print

  if with_labels:
    return (res,totalP,totalN,label)
  else:
    return (res, totalP, totalN, idLabel)


def _computeROC(bn, values, totalP, totalN, idLabel, modalite):
  """
  Parameters
  ----------
  bn : pyAgrum.BayesNet
    a bayesian network
  values :
    the ROC curve values
  totalP : int
    the number of positive values
  totalN : int
    the number of negative values
  idLabel : int
    the id of the target
  modalite :
    the label of the target

  Returns
  -------
  tuple
    (points, opt, seuil)
  """

  res = sorted(values, key=lambda x: x[0])

  vp = 0.0
  fp = 0.0
  xopt = 0.0
  yopt = 0.0
  opt = 100.0
  seuilopt = 0
  points = [(0, 0)]  # first one
  old_seuil = res[0][0]
  for i in range(len(res)):
    # we add a point only if the seuil has changed
    if res[i][0] - old_seuil > 1e-6:  # the seuil allows to take computation variation into account
      x = vp / totalP
      y = fp / totalN
      d = x * x + (1 - y) * (1 - y)
      if d < opt:
        opt = d
        xopt = x
        yopt = y
        seuilopt = old_seuil
      points.append((x, y))
      old_seuil = res[i][0]

    res_id = res[i][1]
    if res_id == idLabel:
      vp += 1.0
    else:
      fp += 1.0

  points.append((1, 1))  # last one

  return points, (xopt, yopt), seuilopt


def module_help(exit_value=1, message=""):
  """
  defines help viewed if args are not OK on command line, and exit with exit_value
  """
  print(os.path.basename(sys.argv[0]), "src.{" + gum.availableBNExts() + "} data[.csv] variable label")
  print()
  print(message)
  print()
  sys.exit(exit_value)


def _drawROC(points, zeTitle, zeFilename, visible, show_fig, save_fig=True,
            special_point=None, special_value=None, special_label=None):
  """
  Draw the ROC curve and save (or not) the curve into zeFilename as a png

  Parameters
  ----------
  points :
    a set of points drawing a ROC curve
  zeTitle : str
    the title of the curve
  zeFilename : str
    the name of the file
  visible : bool
    unnecessary ?
  show_fig : bool
    indicates if the resulting curve must be printed
  save_fig : bool
    indicates if the resulting curve must be saved
  special_point :
    a special point to be highlighted
  special_value :
    a special value to be highlighted
  special_label :
    a special label to be highlighted
  """
  AUC = _computeAUC(points)
  import pylab

  pylab.clf()
  pylab.grid(color='#aaaaaa', linestyle='-', linewidth=1, alpha=0.5)

  pylab.plot([x[0] for x in points], [y[1] for y in points], '-', linewidth=3, color="#000088", zorder=3)
  pylab.fill_between([x[0] for x in points], [y[1] for y in points], 0, color='0.9')
  pylab.plot([0.0, 1.0], [0.0, 1.0], '-', color="#AAAAAA")

  pylab.ylim((-0.01, 1.01))
  pylab.xlim((-0.01, 1.01))
  pylab.xticks(pylab.arange(0, 1.1, .1))
  pylab.yticks(pylab.arange(0, 1.1, .1))
  pylab.grid(True)

  ax = pylab.gca()
  r = pylab.Rectangle((0, 0), 1, 1, edgecolor='#444444', facecolor='none', zorder=1)
  ax.add_patch(r)
  [spine.set_visible(False) for spine in ax.spines.values()]

  if len(points) < 10:
    for i in range(1, len(points) - 1):
      pylab.plot(points[i][0], points[i][1], 'o', color="#000066", zorder=6)

  pylab.xlabel('False positive rate')
  pylab.ylabel('True positive rate')

  if special_point is not None:
    pylab.plot(special_point[0], special_point[1], 'o', color="#DD9999", zorder=6)
    if special_value is not None:
      pylab.text(special_point[0] + 0.01, special_point[1] - 0.01, special_value,
                 {'color': '#DD5555', 'fontsize': 10},
                 horizontalalignment='left',
                 verticalalignment='top',
                 rotation=0,
                 clip_on=False)
  if special_label is not None:
    if special_label != "":
      labels = [special_label]
      colors = ['#DD9999']
      circles = [pylab.Circle((0, 0), 1, fc=colors[0])]
      legend_location = 'lower right'
      pylab.legend(circles, labels, loc=legend_location)

  pylab.text(0.5, 0.3, 'AUC=%f' % AUC,
             horizontalalignment='center',
             verticalalignment='center',
             fontsize=18)

  pylab.title(zeTitle)

  if save_fig:
    pylab.savefig(zeFilename, dpi=300)
    print("\n result in " + zeFilename)

  if show_fig:
    pylab.show()


def showROC(bn, csv_name, variable, label, visible=True, show_fig=False,with_labels=True):
  """
  Compute the ROC curve and save the result in the folder of the csv file.

  Parameters
  ----------
  bn : pyAgrum.BayesNet
    a bayesian network
  csv_name : str
    a csv filename
  target : str
    the target
  label : str
    the target label
  visible : bool
    indicates if the resulting curve must be printed

  """
  (res, totalP, totalN, idTarget) = __computeROCpoints(bn, csv_name, variable, label, visible,with_labels)
  points, opt, seuil = _computeROC(bn, res, totalP, totalN, idTarget, label)

  try:
    shortname = os.path.basename(bn.property("name"))
  except gum.NotFound:
    shortname = "unnamed"

  title = shortname + " vs " + csv_name + " - " + variable + "=" + str(label)

  figname = csv_name + "-" +'ROC_' + shortname + "-" +  variable + "-" + str(label) + '.png'

  _drawROC(points, title, figname, not visible, show_fig, special_point=opt, special_value=seuil)


def _checkROCargs():
  pyAgrum_header("2011-13")

  bn_name = sys.argv[1] if len(sys.argv) > 1 else ""
  csv_name = sys.argv[2] if len(sys.argv) > 2 else ""
  variable = sys.argv[3] if len(sys.argv) > 3 else ""
  label = sys.argv[4] if len(sys.argv) > 4 else ""

  if bn_name.__eq__(""):
    module_help()

  bn = gum.loadBN(bn_name)

  if csv_name.__eq__(""):
    module_help()

  if variable.__eq__(""):
    module_help(message=" Variables : " + str(bn.names()))
  else:
    if variable not in bn.names():
      module_help(message=" Variable '" + variable + "'not found.\n Variables : " + str(bn.names()))

    if label.__eq__(""):
      module_help(message=" Labels : " + str(bn.variableFromName(variable)))
    else:
      try:
        bn.variableFromName(variable)[label]
      except gum.OutOfBounds:
        module_help(message=" Label '" + label + "' not found.\n Labels : " + str(bn.variableFromName(variable)))

  return (bn, csv_name, variable, label)


if __name__ == "__main__":
  pyAgrum_header("2011-17")

  (bn, csv_name, variable, label) = _checkROCargs()
  showROC(bn, csv_name, variable, label, True, False)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# (c) Copyright by Pierre-Henri Wuillemin, UPMC, 2017
# (pierre-henri.wuillemin@lip6.fr)
# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and
# without fee or royalty is hereby granted, provided
# that the above copyright notice appear in all copies
# and that both that copyright notice and this permission
# notice appear in supporting documentation or portions
# thereof, including modifications, that you make.
# THE AUTHOR P.H. WUILLEMIN  DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
# OR PERFORMANCE OF THIS SOFTWARE!

"""
tools for BN analysis in jupyter notebook
"""
from __future__ import print_function

import time

import IPython.display
import base64
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pyAgrum as gum
import pydotplus as dot
from IPython.core.display import Image, display_png
from IPython.core.pylabtools import print_figure
from IPython.display import display, HTML, SVG
from matplotlib.backends.backend_agg import FigureCanvasAgg as fc
from pyAgrum.lib.bn2graph import BN2dot, proba2histo, BNinference2dot, _proba2bgcolor
from pyAgrum.lib.bn_vs_bn import graphicalBNDiff

_cdict = {
  'red'  : ((0.0, 0.1, 0.3),
            (1.0, 0.6, 1.0)),
  'green': ((0.0, 0.0, 0.0),
            (1.0, 0.6, 0.8)),
  'blue' : ((0.0, 0.0, 0.0),
            (1.0, 1, 0.8))
}
_INFOcmap = mpl.colors.LinearSegmentedColormap('my_colormap', _cdict, 256)


def configuration():
  """
  Display the collection of dependance and versions
  """
  from collections import OrderedDict
  import sys, os

  packages = OrderedDict()
  packages["OS"] = "%s [%s]" % (os.name, sys.platform)
  packages["Python"] = sys.version
  packages["IPython"] = IPython.__version__
  packages["MatPlotLib"] = mpl.__version__
  packages["Numpy"] = np.__version__
  packages["pyAgrum"] = gum.__version__

  res = "<table width='100%'><tr><th>Library</th><th>Version</th></tr>"

  for name in packages:
    res += "<tr><td>%s</td><td>%s</td></tr>" % (name, packages[name])

  res += "</table><div align='right'><small>%s</small></div>" % time.strftime(
      '%a %b %d %H:%M:%S %Y %Z')

  display(HTML(res))


def _reprGraph(gr, size, format, asString):
  """
  repr a pydot graph in a notebook

  :param string format : render as png or create_svg
  :param string size : size of the rendered graph
  :param boolean asString : display the graph or return a string containing the corresponding HTML fragment
  """
  gr.set_size(size)
  if format == "svg":
    gsvg = SVG(gr.create_svg())
    if asString:
      return gsvg.data
    else:
      display(HTML(gsvg.data))
  else:
    i = Image(format="png", data=gr.create_png())
    if asString:
      return '<img style="margin:0" src="data:image/png;base64,{}"/>'.format(base64.encodebytes(i.data).decode())
    else:
      display_png(Image(format="png", data=gr.create_png()))


def showGraph(gr, size="4", format="png"):
  """
  show a pydot graph in a notebook

  :param gr: pydot graph
  :param size:  size of the rendered graph
  :param format: render as "png" or "svg"
  :return: the representation of the graph
  """
  return _reprGraph(gr, size, format, asString=False)


def getGraph(gr, size="4", format="png"):
  """
  get a HTML string representation of pydot graph

  :param gr: pydot graph
  :param size:  size of the rendered graph
  :param format: render as "png" or "svg"
  :return: the HTML representation of the graph as a string
  """
  return _reprGraph(gr, size, format, asString=True)


def showDot(dotstring, size="4", format="png"):
  """
  show a dot string as a graph

  :param dotstring: dot string
  :param size: size of the rendered graph
  :param format: render as "png" or "svg"
  :return: the representation of the graph
  """
  return showGraph(dot.graph_from_dot_data(dotstring), size, format)


def getDot(dotstring, size="4", format="png"):
  """
  get a dot string as a HTML string

  :param dotstring: dot string
  :param size: size of the rendered graph
  :param format: render as "png" or "svg"
  :return: the HTML representation of the graph
  """
  g = dot.graph_from_dot_data(dotstring)
  g.set_bgcolor("transparent")
  return getGraph(g, size, format)


def getBNDiff(bn1, bn2, size="4", format="png"):
  """ get a HTML string representation of a graphical diff between the arcs of bn1 (reference) with those of bn2.

  * full black line: the arc is common for both
  * full red line: the arc is common but inverted in bn2
  * dotted black line: the arc is added in bn2
  * dotted red line: the arc is removed in bn2

  :param BayesNet bn1: referent model for the comparison
  :param BayesNet bn2: bn compared to the referent model
  :param size: size of the rendered graph
  :param format: render as "png" or "svg"
  """
  return getGraph(graphicalBNDiff(bn1, bn2), size, format)


def showBNDiff(bn1, bn2, size="4", format="png"):
  """ show a graphical diff between the arcs of bn1 (reference) with those of bn2.

  * full black line: the arc is common for both
  * full red line: the arc is common but inverted in bn2
  * dotted black line: the arc is added in bn2
  * dotted red line: the arc is removed in bn2

  :param BayesNet bn1: referent model for the comparison
  :param BayesNet bn2: bn compared to the referent model
  :param size: size of the rendered graph
  :param format: render as "png" or "svg"
  """
  showGraph(graphicalBNDiff(bn1, bn2), size, format)


def showJunctionTree(bn, withNames=True, size="4", format="png"):
  """
  Show a junction tree

  :param bn: the bayesian network
  :param boolean withNames: display the variable names or the node id in the clique
  :param size: size of the rendered graph
  :param format: render as "png" or "svg"
  :return: the representation of the graph
  """
  jtg = gum.JTGenerator()
  jt = jtg.junctionTree(bn)
  if withNames:
    return showDot(jt.toDotWithNames(bn), size, format)
  else:
    return showDot(jt.toDot(), size, format)


def getJunctionTree(bn, withNames=True, size="4", format="png"):
  """
  get a HTML string for a junction tree (more specifically a join tree)

  :param bn: the bayesian network
  :param boolean withNames: display the variable names or the node id in the clique
  :param size: size of the rendered graph
  :param format: render as "png" or "svg"
  :return: the HTML representation of the graph
  """
  jtg = gum.JTGenerator()
  jt = jtg.junctionTree(bn)
  if withNames:
    return getDot(jt.toDotWithNames(bn), size, format)
  else:
    return getDot(jt.toDot(), size, format)


def showInfluenceDiagram(diag, size="4", format="png"):
  """
  show an influence diagram as a graph

  :param diag: the influence diagram
  :param size: size of the rendered graph
  :param format: render as "png" or "svg"
  :return: the representation of the influence diagram
  """
  return showDot(diag.toDot(), size, format)


def getInfluenceDiagram(diag, size="4", format="png"):
  """
  get a HTML string for an influence diagram as a graph

  :param diag: the influence diagram
  :param size: size of the rendered graph
  :param format: render as "png" or "svg"
  :return: the HTML representation of the influence diagram
  """
  return getDot(diag.toDot(), size, format)


def showProba(p, scale=1.0):
  """
  Show a mono-dim Potential

  :param p: the mono-dim Potential
  :return:
  """
  fig = proba2histo(p, scale)
  plt.show()


def _saveFigProba(p, filename, format="svg"):
  fig = proba2histo(p)
  fig.savefig(filename, bbox_inches='tight', transparent=True, pad_inches=0, dpi=fig.dpi, format=format)
  plt.close(fig)


def getPosterior(bn, evs, target):
  """
  shortcut for getProba(gum.getPosterior(bn,evs,target))

  :param bn: the BayesNet
  :type bn: gum.BayesNet
  :param evs: map of evidence
  :type evs: dict(str->int)
  :param target: name of target variable
  :type target: str
  :return: the matplotlib graph
  """
  return proba2histo(gum.getPosterior(bn, evs=evs, target=target))


def showPosterior(bn, evs, target):
  """
  shortcut for showProba(gum.getPosterior(bn,evs,target))

  :param bn: the BayesNet
  :param evs: map of evidence
  :param target: name of target variable
  """
  showProba(gum.getPosterior(bn, evs=evs, target=target))


def animApproximationScheme(apsc, scale=np.log10):
  """
  show an animated version of an approximation algorithm

  :param apsc: the approximation algorithm
  :param scale: a function to apply to the figure
  """
  from IPython.display import clear_output, display
  f = plt.gcf()

  h = gum.PythonApproximationListener(apsc.asIApproximationSchemeConfiguration())
  apsc.setVerbosity(True)
  apsc.listener = h

  def stopper(x):
    clear_output(True)
    plt.title(
        "{0} \n Time : {1} s | Iterations : {2} | Espilon : {3}".format(x, apsc.currentTime(), apsc.nbrIterations(),
                                                                        apsc.epsilon()))

  def progresser(x, y, z):
    if len(apsc.history()) < 10:
      plt.xlim(1, 10)
    else:
      plt.xlim(1, len(apsc.history()))
    plt.plot(scale(apsc.history()), 'g')
    clear_output(True)
    display(f)

  h.setWhenStop(stopper)
  h.setWhenProgress(progresser)


def showApproximationScheme(apsc, scale=np.log10):
  if apsc.verbosity():
    if len(apsc.history()) < 10:
      plt.xlim(1, 10)
    else:
      plt.xlim(1, len(apsc.history()))
    plt.title(
        "{0} \n Time : {1} s | Iterations : {2} | Espilon : {3}".format(x, apsc.currentTime(), apsc.nbrIterations(),
                                                                        apsc.epsilon()))
    plt.plot(scale(apsc.history()), 'g')
  else:
    ie2.messageApproximationScheme()


def showBN(bn, size="4", format="svg", nodeColor=None, arcWidth=None, arcColor=None, cmap=None, cmapArc=None):
  """
  show a Bayesian network

  :param bn: the bayesian network
  :param size: size of the rendered graph
  :param format: render as "png" or "svg"
  :param nodeColor: a nodeMap of values (between 0 and 1) to be shown as color of nodes (with special colors for 0
  and 1)
  :param arcWidth: a arcMap of values to be shown as width of arcs
  :param arcColor: a arcMap of values (between 0 and 1) to be shown as color of arcs
  :param cmap: color map to show the colors
  :param cmapArc: color map to show the arc color if distinction is needed
  :return: the graph
  """
  if cmapArc is None:
    cmapArc = cmap

  return showGraph(BN2dot(bn, size, nodeColor, arcWidth, arcColor, cmap, cmapArc), size, format)


def getBN(bn, size="4", format="svg", nodeColor=None, arcWidth=None, arcColor=None, cmap=None, cmapArc=None):
  """
  get a HTML string for a Bayesian network

  :param bn: the bayesian network
  :param size: size of the rendered graph
  :param format: render as "png" or "svg"
  :param nodeColor: a nodeMap of values (between 0 and 1) to be shown as color of nodes (with special colors for 0
  and 1)
  :param arcWidth: a arcMap of values to be shown as width of arcs
  :param arcColor: a arcMap of values (between 0 and 1) to be shown as color of arcs
  :param cmap: color map to show the colors
  :param cmapArc: color map to show the arc color if distinction is needed

  :return: the graph
  """
  if cmapArc is None:
    cmapArc = cmap

  return getGraph(BN2dot(bn, size, nodeColor, arcWidth, arcColor, cmap, cmapArc), size, format)


def _normalizeVals(vals, hilightExtrema=False):
  """
  normalisation if vals is not a proba (max>1)
  """
  ma = float(max(vals.values()))
  mi = float(min(vals.values()))
  if ma == mi:
    return None
  else:
    if not hilightExtrema:
      vmi = 0.01
      vma = 0.99
    else:
      vmi = 0
      vma = 1

    try:
      items = vals.items()
    except AttributeError:
      items = vals.iteritems()

    return {name: vmi + (val - mi) * (vma - vmi) / (ma - mi) for name, val in items}


def _reprInformation(bn, evs, size, cmap, asString):
  """
  repr a bn annoted with results from inference : Information and mutual informations

  :param bn: the BN
  :param evs: map of evidence
  :param size:  size of the graph
  :param cmap: colour map used
  :return: the HTML string
  """
  ie = gum.LazyPropagation(bn)
  ie.setEvidence(evs)
  ie.makeInference()

  idEvs = {bn.idFromName(name) for name in evs}
  nodevals = {bn.variable(n).name(): ie.H(n) for n in bn.nodes() if not n in idEvs}
  arcvals = {(x, y): ie.I(x, y) for x, y in bn.arcs()}
  gr = BN2dot(bn, size, nodeColor=_normalizeVals(nodevals, hilightExtrema=False), arcWidth=arcvals, cmapNode=cmap,
              cmapArc=cmap,
              showMsg=nodevals)

  mi = min(nodevals.values())
  ma = max(nodevals.values())

  fig = mpl.figure.Figure(figsize=(8, 3))
  canvas = fc(fig)
  ax1 = fig.add_axes([0.05, 0.80, 0.9, 0.15])
  norm = mpl.colors.Normalize(vmin=mi, vmax=ma)
  cb1 = mpl.colorbar.ColorbarBase(ax1, cmap=cmap,
                                  norm=norm,
                                  orientation='horizontal')
  cb1.set_label('Entropy')
  png = print_figure(canvas.figure, "png")  # from IPython.core.pylabtools
  png_legend = "<img style='vertical-align:middle' src='data:image/png;base64,%s'>" % base64.encodestring(png).decode(
      'ascii')

  gsvg = SVG(gr.create_svg())

  sss = "<div align='center'>" + gsvg.data + "</div>"
  sss += "<div align='center'>"
  sss += "<font color='" + _proba2bgcolor(0.01, cmap) + "'>" + str(mi) + "</font>"
  sss += png_legend
  sss += "<font color='" + _proba2bgcolor(0.99, cmap) + "'>" + str(ma) + "</font>"
  sss += "</div>"

  if asString:
    return sss
  else:
    return display(HTML(sss))


def getInformation(bn, evs=None, size="4", cmap=_INFOcmap):
  """
  get a HTML string for a bn annoted with results from inference : entropy and mutual informations

  :param bn: the BN
  :param evs: map of evidence
  :param size:  size of the graph
  :param cmap: colour map used
  :return: the HTML string
  """
  if evs is None:
    evs = {}

  return _reprInformation(bn, evs, size, cmap, asString=True)


def showInformation(bn, evs=None, size="4", cmap=_INFOcmap):
  """
  show a bn annoted with results from inference : entropy and mutual informations

  :param bn: the BN
  :param evs: map of evidence
  :param size:  size of the graph
  :param cmap: colour map used
  :return: the graph
  """
  if evs is None:
    evs = {}

  return _reprInformation(bn, evs, size, cmap, asString=False)


def showInference(bn, engine=None, evs=None, targets=None, size="7", format='png', nodeColor=None, arcWidth=None,
                  cmap=None):
  """
  show pydot graph for an inference in a notebook

  :param gum.BayesNet bn:
  :param gum.Inference engine: inference algorithm used. If None, LazyPropagation will be used
  :param dictionnary evs: map of evidence
  :param set targets: set of targets
  :param string size: size of the rendered graph
  :param string format: render as "png" or "svg"
  :param boolean asString: display the graph or return a string containing the corresponding HTML fragment
  :return: the desired representation of the inference
  """
  if evs is None:
    evs = {}

  if targets is None:
    targets = {}

  return showGraph(BNinference2dot(bn, size, engine, evs, targets, format, nodeColor, arcWidth, cmap), size, format)


def getInference(bn, engine=None, evs=None, targets=None, size="7", format='png', nodeColor=None, arcWidth=None,
                 cmap=None):
  """
  get a HTML string for an inference in a notebook

  :param gum.BayesNet bn:
  :param gum.Inference engine: inference algorithm used. If None, LazyPropagation will be used
  :param dictionnary evs: map of evidence
  :param set targets: set of targets
  :param string size: size of the rendered graph
  :param string format: render as "png" or "svg"
  :param nodeColor: a nodeMap of values (between 0 and 1) to be shown as color of nodes (with special colors for 0
  and 1)
  :param arcWidth: a arcMap of values to be shown as width of arcs
  :param cmap: color map to show the color of nodes and arcs
  :return: the desired representation of the inference
  """
  if evs is None:
    evs = {}

  if targets is None:
    targets = {}

  return getGraph(BNinference2dot(bn, size, engine, evs, targets, format, nodeColor, arcWidth, cmap), size, format)


def _reprPotential(pot, digits=4, withColors=True, varnames=None, asString=False):
  """
  return a representation of a gum.Potential as a HTML table.
  The first dimension is special (horizontal) due to the representation of conditional probability table


  :param pot: the potential to get
  :param digits: number of digits to show
  :param: withColors : bgcolor for proba cells or not
  :param varnames: the aliases for variables name in the table
  :param asString: display the table or a HTML string
  :return: the representation
  """
  from IPython.core.display import HTML

  def _rgb(r, g, b):
    return '#%02x%02x%02x' % (r, g, b)

  def _mkCell(val):
    s = "<td style='"
    if withColors and (0 <= val <= 1):
      r = int(255 - val * 128)
      g = int(127 + val * 128)
      b = 100
      s += "background-color:" + _rgb(r, g, b) + ";"
    s += "text-align:right;'>{:." + str(digits) + "f}</td>"
    return s.format(val)

  html = list()
  html.append("<table>")
  if pot.empty():
    html.append("<tr><th style='background-color:#AAAAAA'>&nbsp;</th></tr>")
    html.append("<tr>" + _mkCell(pot.get(gum.Instantiation())) + "</tr>")
  else:
    if varnames is not None and len(varnames) != pot.nbrDim():
      raise ValueError(
          "varnames contains {} values instead of the needed {} values.".format(len(varnames), pot.nbrDim()))

    nparents = pot.nbrDim() - 1
    var = pot.variable(0)
    varname = var.name() if varnames == None else varnames[0]

    # first line
    if nparents > 0:
      html.append(
          "<tr><th colspan='{}'></th><th colspan='{}' style='background-color:#AAAAAA'><center>{"
          "}</center></th></tr>".format(
              nparents, var.domainSize(), varname))
    else:
      html.append(
          "<tr style='background-color:#AAAAAA'><th colspan='{}'><center>{}</center></th></tr>".format(
              var.domainSize(),
              varname))
    # second line
    s = "<tr>"
    if nparents > 0:
      # for parent in pot.var_names[:-1] if varnames == None else varnames[1:]:
      for par in range(nparents - 1, 0 - 1, -1):
        parent = pot.var_names[par] if varnames is None else varnames[par]
        s += "<th style='background-color:#AAAAAA'><center>{}</center></th>".format(parent)
    for label in var.labels():
      s += "<th style='background-color:#BBBBBB'><center>{}</center></th>".format(label)
    s += '</tr>'

    html.append(s)

    inst = gum.Instantiation(pot)
    off = 1
    offset = dict()
    for i in range(1, nparents + 1):
      offset[i] = off
      off *= inst.variable(i).domainSize()

    inst.setFirst()
    while not inst.end():
      s = "<tr>"
      for par in range(1, nparents + 1):
        label = inst.variable(par).label(inst.val(par))
        if par == 1:
          s += "<th style='background-color:#BBBBBB'><center>{}</center></th>".format(label)
        else:
          if sum([inst.val(i) for i in range(1, par)]) == 0:
            s += "<th style='background-color:#BBBBBB;' rowspan = '{}'><center>{}</center></th>".format(offset[par],
                                                                                                        label)
      for j in range(pot.variable(0).domainSize()):
        s += _mkCell(pot.get(inst))
        inst.inc()
      s += "</tr>"
      html.append(s)

  html.append("</table>")

  if asString:
    return "\n".join(html)
  else:
    return HTML("".join(html))


def __isKindOfProba(pot):
  """
  check if pot is a joint proba or a CPT
  :param pot: the potential
  :return: True or False
  """
  if abs(pot.sum() - 1) < 1e-2:
    return True

  q = pot.margSumOut([pot.variable(0).name()])
  if abs(q.max() - 1) > 1e-2:
    return False
  if abs(q.min() - 1) > 1e-2:
    return False
  return True


def showPotential(pot, digits=4, withColors=None, varnames=None):
  """
  show a gum.Potential as a HTML table.
  The first dimension is special (horizontal) due to the representation of conditional probability table

  :param gum.Potential pot: the potential to get
  :param int digits: number of digits to show
  :param: boolean withColors : bgcolor for proba cells or not
  :param list of strings varnames: the aliases for variables name in the table
  :return: the display of the potential
  """
  if withColors is None:
    withColors = __isKindOfProba

  display(_reprPotential(pot, digits, withColors, varnames, asString=False))


def getPotential(pot, digits=4, withColors=None, varnames=None):
  """
  return a HTML string of a gum.Potential as a HTML table.
  The first dimension is special (horizontal) due to the representation of conditional probability table

  :param gum.Potential pot: the potential to get
  :param int digits: number of digits to show
  :param: boolean withColors : bgcolor for proba cells or not
  :param list of strings varnames: the aliases for variables name in the table
  :return: the HTML string
  """
  if withColors is None:
    withColors = __isKindOfProba(pot)

  return _reprPotential(pot, digits, withColors, varnames, asString=True)


def getSideBySide(*args, **kwargs):
  """
  create an HTML table for args as string (using string, _repr_html_() or str())

  :param args: HMTL fragments as string arg, arg._repr_html_() or str(arg)
  :param captions: list of strings (captions)
  :return: a string representing the table
  """

  if not set(kwargs.keys()).issubset(set(['captions', 'valign'])):
    raise TypeError("sideBySide() got unexpected keyword argument(s) : '{}'".format(
        set(kwargs.keys()).difference(set(['captions', 'valign']))))

  if 'captions' in kwargs:
    captions = kwargs['captions']
  else:
    captions = None

  if 'valign' in kwargs:
    v_align = 'vertical-align:' + kwargs['valign'] + ';'
  else:
    v_align = ""

  s = '<table style="border-style: hidden; border-collapse: collapse;" width="100%">'

  def reprHTML(s):
    if isinstance(s, str):
      return s
    elif hasattr(s, '_repr_html_'):
      return s._repr_html_()
    else:
      return str(s)

  s += '<tr><td style="border-top:hidden;border-bottom:hidden;' + v_align + '"><div align="center" style="' + v_align + '">'
  s += ('</div></td><td style="border-top:hidden;border-bottom:hidden;' + v_align + '"><div align="center" style="' + v_align + '">').join([reprHTML(arg)
                                                                                                   for arg in args])
  s += '</div></td></tr>'

  if captions is not None:
    s += '<tr><td style="border-top:hidden;border-bottom:hidden;"><div align="center"><small>'
    s += '</small></div></td><td style="border-top:hidden;border-bottom:hidden;"><div align="center"><small>'.join(
        captions)
    s += '</small></div></td></tr>'

  s += '</table>'
  return s


def sideBySide(*args, **kwargs):
  """
  display side by side args as HMTL fragment (using string, _repr_html_() or str())

  :param args: HMTL fragments as string arg, arg._repr_html_() or str(arg)
  :param captions: list of strings (captions)
  """
  display(HTML(getSideBySide(*args, **kwargs)))


def getInferenceEngine(ie, inferenceCaption):
  """
  display an inference as a BN+ lists of hard/soft evidence and list of targets

  :param gum.InferenceEngine ie: inference engine
  :param string inferenceCaption: title for caption

  """
  t = '<div align="left"><ul>'
  if ie.nbrHardEvidence() > 0:
    t += "<li><b>hard evidence</b><br/>"
    t += ", ".join([ie.BN().variable(n).name() for n in ie.hardEvidenceNodes()])
    t += "</li>"
  if ie.nbrSoftEvidence() > 0:
    t += "<li><b>soft evidence</b><br/>"
    t += ", ".join([ie.BN().variable(n).name() for n in ie.softEvidenceNodes()])
    t += "</li>"
  if ie.nbrTargets() > 0:
    t += "<li><b>target(s)</b><br/>"
    if ie.nbrTargets() == ie.BN().size():
      t += " all"
    else:
      t += ", ".join([ie.BN().variable(n).name() for n in ie.targets()])
    t += "</li>"

  if hasattr(ie, 'nbrJointTargets'):
    if ie.nbrJointTargets() > 0:
      t += "<li><b>Joint target(s)</b><br/>"
      t += ", ".join(['['
                      + (", ".join([ie.BN().variable(n).name() for n in ns]))
                      + ']' for ns in ie.jointTargets()])
      t += "</li>"
  t += '</ul></div>'
  return getSideBySide(getBN(ie.BN()), t, captions=[inferenceCaption, "Evidence and targets"])


# adding _repr_html_ to some pyAgrum classes !
gum.BayesNet._repr_html_ = lambda self: getBN(self)
gum.Potential._repr_html_ = lambda self: getPotential(self)
gum.LazyPropagation._repr_html_ = lambda self: getInferenceEngine(self, "Lazy Propagation on this BN")

gum.UndiGraph._repr_html_ = lambda self: getDot(self.toDot())
gum.DiGraph._repr_html_ = lambda self: getDot(self.toDot())
gum.MixedGraph._repr_html_ = lambda self: getDot(self.toDot())
gum.DAG._repr_html_ = lambda self: getDot(self.toDot())
gum.CliqueGraph._repr_html_ = lambda self: getDot(self.toDot())
gum.EssentialGraph._repr_html_ = lambda self: getDot(self.toDot())
gum.MarkovBlanket._repr_html_ = lambda self: getDot(self.toDot())
